#!/usr/bin/env ruby

require "optparse"
require "ostruct"
require "xmlrpc/client"

# A couple things that still need to be done:
#   - Add a quiet mode
#   - Option for the event builders to sprout a canvas
#   - Make disk writing optional
#   - Validate configurations looking for simple problems
#   - Calculate the optimum number of buffers for each process

# It'd be nice if we were using a newer version of ruby (1.9.3) that had
# better string format substitution.  For the time being we'll setup our
# string constants using the newer convention and do the ugly substitutions.
EB_DISK_CONFIG = "\
daq: {
  max_fragment_size_words: 524288
  event_builder: {
    mpi_buffer_count: %{buffer_count}
    first_fragment_receiver_rank: 0
    fragment_receiver_count: %{total_frs}
    expected_fragments_per_event: %{total_frs}
    use_art: true
    print_event_store_stats: true
  }
}
physics: {
  my_output_modules: [ normalOutput ]
  end_paths: [ my_output_modules ]
}
outputs: {
  normalOutput: {
    module_type: RootOutput
    fileName: \"%{output_file}\"
    compressionLevel: 0
  }
}
source: {
  module_type: RawInput
  waiting_time: 20
  resume_after_timeout: true
  fragment_type_map: [[1, \"missed\"], [2, \"V1495\"], [3, \"V1720\"], [4, \"V1724\"]]
}
process_name: ONLINEDEMO"

V1495_CONFIG = "\
daq: {
  max_fragment_size_words: 524288
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V1495Driver
    link_type: PCIE
    link_number: %{link_number}
    vme_base: 0x01000000
    pulser_frequency: %{frequency}
    acquisition_gate_us: 410
    memory_full_extension_us: 50
    low_threshold: 0
    generator_ds50: {
      fragment_id: %{last_fragment_id}
    }
  }
}"

V1720_CONFIG = "\
BEGIN_PROLOG
def_offset: 0x2700
def_channels: [ { channel: 0 enable: true dc_offset: @local::def_offset },
                { channel: 1 enable: true dc_offset: @local::def_offset },
                { channel: 2 enable: true dc_offset: @local::def_offset },
                { channel: 3 enable: true dc_offset: @local::def_offset },
                { channel: 4 enable: true dc_offset: @local::def_offset },
                { channel: 5 enable: true dc_offset: @local::def_offset },
                { channel: 6 enable: true dc_offset: @local::def_offset },
                { channel: 7 enable: true dc_offset: @local::def_offset } ]
END_PROLOG
daq: {
  max_fragment_size_words: 524288
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V172xDriver
    link_type: PCIE
    link_number: %{link_number}
    io_level: TTL
    acquisition_gate_us: 400
    pre_trigger_us: 40
    acquisition_mode: HW
    multiboard_sincronicazion: true
    channels: @local::def_channels
    generator_ds50: {
      fragment_id: %{fragment_id}
    }
  }
}"

V1720_SIM_CONFIG = "\
daq: {
  max_fragment_size_words: 524288 
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    mpi_sync_interval: 50
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V172xSimulator 
    freqs_file: \"ds50_hist.dat\" 
    fragments_per_event: 1 
    nChannels: 3000 
    generator_ds50: {
      fragment_id: %{fragment_id}
    }
  }
}"

class ConfigGen
  def generateEventBuilder(ebIndex, totalFRs, dataDir, runNumber)
    # Do the substitutions in the event builder configuration given the options
    # that were passed in from the command line.  Assure that files written out
    # by each EB are unique by including a timestamp in the file name.
    ebConfig = String.new(EB_DISK_CONFIG)
    ebConfig.gsub!(/\%\{total_frs\}/, String(totalFRs))
    ebConfig.gsub!(/\%\{buffer_count\}/, String(totalFRs*10))
    currentTime = Time.now
    fileName = "ds50daq_run%06d_eb%02d_" % [Integer(runNumber), ebIndex]
    fileName += "%d%02d%02d-%02d%02d%02d" % [currentTime.year, currentTime.month,
                                             currentTime.day, currentTime.hour,
                                             currentTime.min, currentTime.sec]
    fileName += ".root"
    outputFile = File.join(dataDir, fileName)
    ebConfig.gsub!(/\%\{output_file\}/, outputFile)
    return ebConfig
  end
  
  def generateV1495(totalEBs, totalFRs, linkNumber, frequency)
    # Do the substitutions in the V1495 configuration given the options that 
    # were passed in from the command line.
    v1495Config = String.new(V1495_CONFIG)
    v1495Config.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    v1495Config.gsub!(/\%\{total_frs\}/, String(totalFRs))
    v1495Config.gsub!(/\%\{last_fragment_id\}/, String(totalFRs-1))
    v1495Config.gsub!(/\%\{link_number\}/, String(linkNumber))
    v1495Config.gsub!(/\%\{frequency\}/, String(frequency))
    v1495Config.gsub!(/\%\{buffer_count\}/, String(totalEBs*10))
    return v1495Config
  end
  
  def generateV1720(v1720Index, totalEBs, totalFRs, linkNumber, sim)
    # Do the substitutions in the V1720 configuration given the options that 
    # were passed in from the command line.
    if sim != 0
      v1720Config = String.new(V1720_SIM_CONFIG)
    else
      v1720Config = String.new(V1720_CONFIG)
    end
    
    v1720Config.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    v1720Config.gsub!(/\%\{total_frs\}/, String(totalFRs))
    v1720Config.gsub!(/\%\{link_number\}/, String(linkNumber))
    v1720Config.gsub!(/\%\{fragment_id\}/, String(v1720Index))
    v1720Config.gsub!(/\%\{buffer_count\}/, String(totalEBs*10))
    return v1720Config
  end
end

class CommandLineParser
  def initialize(configGen)
    @configGen = configGen
    @options = OpenStruct.new
    @options.eventBuilders = []
    @options.v1720s = []
    @options.v1495s = []
    @options.dataDir = nil
    @options.command = nil
    @options.summary = false
    @options.runNumber = "0101"
    @options.serialize = false

    @optParser = OptionParser.new do |opts|
      opts.banner = "Usage: ds50MasterCtrl.rb [options]"
      opts.separator ""
      opts.separator "Specific options:"

      opts.on("--eb [host,port]", Array, "Add an event builder that runs on the",
              "specified host and port.") do |eb|
        ebConfig = OpenStruct.new
        ebConfig.host = eb[0]
        ebConfig.port = Integer(eb[1])
        ebConfig.kind = "eb"
        @options.eventBuilders << ebConfig
      end
    
      opts.on("--v1495 [host,port,link_number,frequency]", Array, 
              "Add a V1495 fragment receiver that runs on",
              "the specified host, port, and link", 
              "generating pulses at the given frequency.") do |v1495|
        if v1495.length != 4
          puts "You must specifiy a host, port, link_number and frequency"
          puts "for the V1495."
          exit
        end
        v1495Config = OpenStruct.new
        v1495Config.host = v1495[0]
        v1495Config.port = Integer(v1495[1])
        v1495Config.link_number = Integer(v1495[2])
        v1495Config.pulser_frequency = Integer(v1495[3])
        v1495Config.kind = "v1495"
        @options.v1495s << v1495Config
      end
    
      opts.on("--v1720 [host,port,link_number,sim]", Array, 
              "Add a V1720 fragment receiver that runs on",
              "the specified host, port and link.  The sim", 
              "parameter (0 or 1) determines whether or",
              "not to use the V1720 simulator.") do |v1720|
        if v1720.length != 4
          puts "You must specifiy a host, port, link_number and whether or not"
          puts "the V1720 is simulated."
          exit
        end
        v1720Config = OpenStruct.new
        v1720Config.host = v1720[0]
        v1720Config.port = Integer(v1720[1])
        v1720Config.link_number = Integer(v1720[2])
        v1720Config.sim = Integer(v1720[3])
        v1720Config.kind = "v1720"
        @options.v1720s << v1720Config
      end

      opts.on("-d", "--data-dir [data dir]", 
              "Directory that the event builders will", "write data to.") do |dataDir|
        @options.dataDir = dataDir
      end

      opts.on("-c", "--command [command]", 
              "Execute a command: start, stop, init.") do |command|
        @options.command = command
      end

      opts.on("-r", "--run-number [number]", "Specify the run number.") do |run|
        @options.runNumber = run
      end

      opts.on("-s", "--summary", "Summarize the configuration.") do
        @options.summary = true
      end

      opts.on("-e", "--serialize", "Serialize the config for System Control.") do
        @options.serialize = true
      end

      opts.on_tail("-h", "--help", "Show this message.") do
        puts opts
        exit
      end
    end
  end

  def parse()
    if ARGV.length == 0
      puts @optParser
      exit
    end

    @optParser.parse(ARGV)
    self.validate()
  end

  def validate()
    # Make sure link numbers are unique
    # In data mode make sure there is an eb, 1720 and 1495
    # In sim mode make sure there is an eb and a 1720
    return nil
  end

  def summarize()
    # Print out a summary of the configuration that was passed in from the
    # the command line.  Everything will be printed in terms of what process
    # is running on which host.
    puts "Configuration Summary:"
    hostMap = {}
    (@options.eventBuilders + @options.v1495s + @options.v1720s).each do |proc|
      if not hostMap.keys.include?(proc.host)
        hostMap[proc.host] = []
      end
      hostMap[proc.host] << proc
    end

    hostMap.each_key { |host|
      puts "  %s:" % [host]
      hostMap[host].sort! { |x,y|
        x.port <=> y.port
      }
      hostMap[host].each do |item|
        case item.kind
        when "eb"
          puts "    EventBuilder, port %d" % item.port
        when "v1495"
          puts "    FragmentReceiver, V1495, port %s, link %s" % [item.port, 
                                                                  item.link_number]
        when "v1720"
          if item.sim != 0
            puts "    FragmentReceiver, Simulated V1720, port %d" % [item.port]
          else
            puts "    FragmentReceiver, V1720, port %d, link %d" % [item.port, 
                                                                    item.link_number]
          end
        end
      end
      puts ""
    }
    return nil
  end

  def serialize()
    # The current system control pulls the FHCL config to initialize DAQ
    # processes from a series of files on disk.  The files are named as follows:
    #  initorder#hostname#port#description#on.fcl
    # 
    # This method will generate configurations given the command line options
    # and write out files for System Control.
    puts "Serializing config:"
    procIndex = 1
    v1720Index = 0
    ebIndex = 0
    totalFRs = @options.v1495s.length + @options.v1720s.length
    totalEBs = @options.eventBuilders.length
    desc = "%dx%d" % [totalFRs, totalEBs]
    (@options.eventBuilders + @options.v1495s + @options.v1720s).each do |proc|
      fileName = "%d#%s#%d#%s#on.fcl" % [procIndex, proc.host, proc.port, desc]
      procIndex += 1
      puts "  writing %s..." % fileName
      handle = File.open(fileName, "w")
      case proc.kind
      when "eb"
        cfg = @configGen.generateEventBuilder(ebIndex, totalFRs, @options.dataDir,
                                              @options.runNumber)
        ebIndex += 1
      when "v1495"
        cfg = @configGen.generateV1495(totalEBs, totalFRs, proc.link_number,
                                       proc.pulser_frequency)
      when "v1720"
        cfg = @configGen.generateV1720(v1720Index, totalEBs, totalFRs, 
                                       proc.link_number, proc.sim)
        v1720Index += 1
      end
      handle.write(cfg)
      handle.close()
    end
  end

  def getOptions()
    return @options
  end
end

class SystemControl
  def initialize(cmdLineOptions, configGen)
    @options = cmdLineOptions
    @configGen = configGen
  end

  def init()
    # The DAQ components need to be initialized in parallel as this is the step
    # in which connections are made between the various components.  The XMLRPC
    # call will block until all connections have been made which means that the
    # configurations need to be posted to all processes at the same time.
    threads = []
    ebIndex = 0
    totalFRs = @options.v1495s.length + @options.v1720s.length
    totalEBs = @options.eventBuilders.length
    @options.eventBuilders.each { |ebOptions|
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(ebOptions.host, "/RPC2", 
                                          ebOptions.port)
        cfg = @configGen.generateEventBuilder(ebIndex, totalFRs, @options.dataDir,
                                              @options.runNumber)
        result = xmlrpcClient.call("daq.init", cfg)
        puts "EventBuilder on %s:%d result: %s" % [ebOptions.host, ebOptions.port,
                                                   result]
      end
      ebIndex += 1
    }
    
    @options.v1495s.each { |v1495Options|
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(v1495Options.host, "/RPC2", 
                                          v1495Options.port)
        cfg = @configGen.generateV1495(totalEBs, totalFRs, 
                                       v1495Options.link_number,
                                       v1495Options.pulser_frequency)
        result = xmlrpcClient.call("daq.init", cfg)
        puts "V1495 FragmentReceiver on %s:%d result: %s" % [v1495Options.host, 
                                                             v1495Options.port,
                                                             result]
      end
    }
    v1720Index = 0
    @options.v1720s.each { |v1720Options|
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(v1720Options.host, "/RPC2",
                                          v1720Options.port)
        cfg = @configGen.generateV1720(v1720Index, totalEBs, totalFRs, 
                                       v1720Options.link_number,
                                       v1720Options.sim)
        result = xmlrpcClient.call("daq.init", cfg)
        puts "V1720 FragmentReceiver on %s:%d result: %s" % [v1720Options.host, 
                                                             v1720Options.port,
                                                             result]
      end
      v1720Index += 1      
    }
    
    threads.each { |aThread|
      aThread.join()
    }
  end

  def start(runNumber)
    # Starting runs is simple in that all processes are sent the same message
    # and that can be done serially.
    (@options.eventBuilders + @options.v1495s + @options.v1720s).each do |proc|
      puts "Attempting to connect to %s:%d and start a run." % [proc.host, 
                                                                proc.port]
      xmlrpcClient = XMLRPC::Client.new(proc.host, "/RPC2", proc.port)
      result = xmlrpcClient.call("daq.start", runNumber)
      case proc.kind
      when "eb"
        puts "EventBuilder on %s:%d result: %s" % [proc.host, proc.port, result]
      when "v1495"
        puts "V1495 FragmentReceiver on %s:%d result: %s" % [proc.host, proc.port, 
                                                             result]
      when "v1720"
        puts "V1720 FragmentReceiver on %s:%d result: %s" % [proc.host, proc.port,
                                                             result]
      end
    end
  end

  def stop
    threads = []
    (@options.v1495s + @options.v1720s + @options.eventBuilders).each do |proc|
      threads << Thread.new do
        puts "Attempting to connect to %s:%d and stop a run." % [proc.host, 
                                                                 proc.port]
        xmlrpcClient = XMLRPC::Client.new(proc.host, "/RPC2", proc.port)
        result = xmlrpcClient.call("daq.stop")
        case proc.kind
        when "eb"
          puts "EventBuilder on %s:%d result: %s" % [proc.host, proc.port, 
                                                     result]
        when "v1495"
          puts "V1495 FragmentReceiver on %s:%d result: %s" % [proc.host, proc.port, 
                                                               result]
        when "v1720"
          puts "V1720 FragmentReceiver on %s:%d result: %s" % [proc.host, proc.port,
                                                               result]
        end
      end
    end
    threads.each { |aThread|
      aThread.join
    }
  end
end

if __FILE__ == $0
  cfgGen = ConfigGen.new
  cmdLineParser = CommandLineParser.new(cfgGen)
  cmdLineParser.parse()
  options = cmdLineParser.getOptions()

  if options.summary
    cmdLineParser.summarize()
  end

  if options.serialize
    cmdLineParser.serialize()
    exit
  end

  sysCtrl = SystemControl.new(options, cfgGen)

  if options.command == "init"
    sysCtrl.init()
  elsif options.command == "start"
    sysCtrl.start(options.runNumber)
  elsif options.command == "stop"
    sysCtrl.stop()
  end
end
